"""Submodule providing edge prediction models based on Sklearn Models."""
from .sklearn_model_edge_prediction_adapter import SklearnModelEdgePredictionAdapter

__all__ = [
    "SklearnModelEdgePredictionAdapter"
]