"""Submodule providing node-label prediction models based on Sklearn Models."""

from .sklearn_model_node_label_prediction_adapter import SklearnModelNodeLabelPredictionAdapter

__all__ = [
    "SklearnModelNodeLabelPredictionAdapter"
]
