"""Submodule providing edge-label prediction models based on Sklearn Models."""

from .sklearn_model_edge_label_prediction_adapter import SklearnModelEdgeLabelPredictionAdapter

__all__ = [
    "SklearnModelEdgeLabelPredictionAdapter"
]