"""Submodule providing models for edge-label prediction."""
from .edge_label_prediction_sklearn import SklearnModelEdgeLabelPredictionAdapter

__all__ = [
    "SklearnModelEdgeLabelPredictionAdapter"
]